//<c>   Copyright 1993-1999 by Gerry J. Danen; all rights reserved.

//<t>   Audit NetWare Passwords

//      at start, turn intruder checking off, at end, turn it back on
//      must be supervisor-equivalent to run
//      loop thru bindery, using VerifyBinderyObjectPassword()

//      to do:
//
//      /n - check non-users (! OT_USER)
//      /b=name[word] - begin checking at name[word]
//      /d=date - check only passwords changed since date
//      /l - set users with offending passwords to "locked out" status (disable account)
//      /s - show offending passwords in secret file
//      /u - unsorted bindery processing
//
//      check login control record in bindery for minimum req'd length, and
//      other potential security problems
//
//      allow looping by lookup, then user (if new words are at the top of the
//      list)
//
//      change users with guessed passwords to 01-Jan-85 expiration date
//
//      bypass users with expired passwords (<= today's date)

#define IS_AUDIT_PW
#include"nss.h"


static  NWVOL_USAGE vu ;
static  t_volume_info vi[32] ;
static  WORD    maxVolume, serverConnID ;
static  char    buf[BUFSIZE], buf1[BUFSIZE], kb_pass[50]="" ;
static  UCHAR   serverName[SERVERNAME_SIZE], userName[B_USERNAME_SIZE], startName[B_USERNAME_SIZE] ;
static  UCHAR   datafile[FS_LEN]="",
                statusfile[FS_LEN]="$audit.sf",
                researchfile[FS_LEN]="$audit.ui",
                pwfile[FS_LEN]="" ;
static  FILE    *rfile, *sfile ;
static  BOOL    disable_account=FALSE, research=TRUE, restart=FALSE, show_pw=FALSE, write_status=TRUE, fl_unsorted=FALSE, keyboard_locked=FALSE ;
static  ULONG   checks=0, progress=0 ;
static  int     lookup_end=0, user_end=0, risks=0, total_users=0, this_user=0, first_line_to_scroll=8 ;
static  char    *lookup[MAXLOOK+1], *users[MAXUSER+1] ;
static  STRING  exitmsg[] =
                {
                (STRING) "",
                (STRING) "Cannot open data file",
                (STRING) "Cannot allocate enough memory to hold data file",
                (STRING) "Server out of memory",
                (STRING) "Intruder detection is activeturn it off first with SYSCON",
                (STRING) "Cannot allocate enough memory to hold user list",
                (STRING) ""
                };
static  INT32   sys_ymd, sys_ymdn ;
static  UCHAR   flpart1[]="stront" ;            // makes part of a file name


#include"nss.ccc"


void    do_help( void )
{
    fprintf( stderr, "\n\n\nUsage:\n\t%s [/b=startname] ...\n\n", get_global_program_name() );

    exit( GD_ERROR_HELP );
}


int     qs_compare( const char **arg1, const char **arg2 )
{
    return( _stricmp( *arg1, *arg2 ) );
}


STRING  system_dir( void )
{
    return( "system" );
}


void    pr_centered( int l, STRING m )
{
    DspMsg( l, CenterMsg(m), vid.err_fg, vid.err_bg, m, YES );
}


void    topbox( STRING msg )
{
    box( vid.err_fg, vid.err_bg, 5,1, 7,vid.cols, YES, 1 );
    pr_centered( 6, msg );
}


void    pr( int lf_before, int lf_after, STRING fmt, ... )
{
    static int  l ;
    va_list     arg_ptr ;

    va_start( arg_ptr, fmt );
    vsprintf( ErrMsg_buf_, fmt, arg_ptr );
    va_end( arg_ptr );
    if ( lf_before > 0 )
        sc_scroll( lf_before, DEFAULT_ATTR, first_line_to_scroll, 1, vid.rows-1, vid.cols, SCROLL_UP );
    DspMsg( vid.rows-1, 1, vid.normal_fg, vid.normal_bg, ErrMsg_buf_, YES );
    if ( lf_after > 0 )
        sc_scroll( lf_after, DEFAULT_ATTR, first_line_to_scroll, 1, vid.rows-1, vid.cols, SCROLL_UP );
}


void    enter_kb_pw( void )
{
    int         ch, x=0, szDest ;
    UCHAR       s[50] ;

    if ( keyboard_locked )
        topbox( "Password to unlock the keyboard?" );
    else
        topbox( "Enter password to lock the keyboard" );

    s[0] = EOS ;
    szDest = sizeof(s)-2 ;
    do  {
        ch = kb_getkey_();
        if ( ch > ' ' && ch < '~' )
        {
            if ( x < szDest )
            {
                s[x] = (UCHAR) ch ;
                x += 1 ;
                s[x] = EOS ;
            }
        }
        if ( ch == BACKSP )
        {
            if ( x < 1 )
                sn_beep();
            else
            {
                x -= 1 ;
                s[x] = EOS ;
            }
        }
    }   while ( ch != RETURN );
    st_trim( s );
    strupr ( s );
    if ( keyboard_locked )
    {
        if ( st_eq(s,kb_pass) )
        {
            keyboard_locked = FALSE ;
            topbox( "Keyboard is unlocked" );
            blank_message = "Password audit in progress" ;
        }
        else
            topbox( "Wrong password.  Keyboard is still locked!" );
    }
    else
    {
        if ( strlen(s) )
        {
            strcpy( kb_pass, s );
            keyboard_locked = TRUE ;
            topbox( "Keyboard is now locked.  Remember the password!" );
            blank_message = "Password audit in progress -- keyboard is locked" ;
        }
        else
            topbox( "No password.  Keyboard is NOT locked!" );
    }
}


void    upd_statusfile( STRING fmt, ... )
{
    FILE        *fh ;
    va_list     arg_ptr ;

    va_start( arg_ptr, fmt );
    vsprintf( ErrMsg_buf_, fmt, arg_ptr );
    va_end( arg_ptr );

    fh = fopen( statusfile, "a+" );
    if ( fh )
    {
        fprintf( fh, "%s %s %s %s\n", TodaysDay(), TodaysDate(), TodaysTime(), ErrMsg_buf_ );
        fclose ( fh );
    }
}


void    append_to_secret_file( STRING username, STRING userpw )
{
#if IS_USED_INTERNAL_ONLY
    UCHAR       fn[FS_LEN], dot='.', au[]="audit" ;
    FILE        *fh ;

    sprintf( pwfile, "%s%s\\%s", SYSTEM_VOLUME, system_dir(), au );
    dr_mkdir( pwfile, NO );
    sprintf( fn, "%s\\%s%c%s", pwfile, flpart1, dot, "GjD" );
    fh = fopen( fn, "a+" );
    if ( fh )
    {
        fprintf( fh, "%s %s %s %s %s\n", TodaysDay(), TodaysDate(), TodaysTime(), username, userpw );
        fclose ( fh );
    }
#endif
}


void    research_login_data( char oName[], WORD oType )
{
    int         x, ccode ;
    ULONG       oID, volRestriction, volInUse, totalInUse ;
    BOOL        isRestr ;

    ccode = get_login_control( oName, oType, &oID, propValue, &lt );
    if ( ccode )
    {
        fprintf( rfile, "\n*** ReadPropertyValue error %04Xh *** user %s", ccode, oName );
    }
    else
    {
        fprintf( rfile, "\nAccount Information for User %s [%08lX]", oName, oID );
        nw_formatted_bindery_date( &lt.ac_expires, buf );
        fprintf( rfile, "\nAccount Disabled:                 %s", (lt.acct_disabled == 0xFF ? "YES" : "No") );
        fprintf( rfile, "\nAccount Has Expiration Date:      %s", st_eq(buf,"-") ? "No" : "Yes" );
        fprintf( rfile, "\n   Date Account Expires:          %s", buf );
        fprintf( rfile, "\nLimit Concurrent Connections:     %s", (lt.sim_connect.n == 0 ? "No" : "Yes") );
        fprintf( rfile, "\n   Maximum Connections:           " );
        if ( lt.sim_connect.n > 0 )
            fprintf( rfile, "%d", lt.sim_connect.n );
        else
            fprintf( rfile, "unlimited" );
        fprintf( rfile, "\nAllow User To Change Password:    !!" );
        fprintf( rfile, "\nRequire Password:                 " );
        if ( lt.min_pw_length == 0 &&
             lt.days_betw_pw_changes.n == 0 &&
             (lt.pw_expires.y==0 && lt.pw_expires.m==0 && lt.pw_expires.d==0) &&
             (lt.grace_allowed == 0xFF && lt.grace_left == 0xFF) )
            fprintf( rfile, "No" );
        else
            fprintf( rfile, "Yes" );
        fprintf( rfile, "\n   Minimum Password Length:       %d", lt.min_pw_length );
        fprintf( rfile, "\nForce Periodic Password Changes:  !!" );
        fprintf( rfile, "\n   Days Between Forced Changes:   %d", lt.days_betw_pw_changes.n );
        fprintf( rfile, "\n   Date Password Expires:         %s", nw_formatted_bindery_date(&lt.pw_expires, buf) );
        fprintf( rfile, "\n   Limit Grace Logins:            %s", (lt.grace_allowed == 0xFF || lt.grace_left == 0xFF ? "No" : "Yes") );
        fprintf( rfile, "\n      Grace Logins Allowed:       " );
        if ( lt.grace_allowed == 0xFF )
            fprintf( rfile, "unlimited" );
        else
            fprintf( rfile, "%d", lt.grace_allowed );
        fprintf( rfile, "\n      Remaining Grace Logins:     " );
        if ( lt.grace_left == 0xFF )
            fprintf( rfile, "unlimited" );
        else
            fprintf( rfile, "%d", lt.grace_left );
        fprintf( rfile, "\nRequire Unique Passwords:         !!" );
        nw_formatted_bindery_date( &lt.last_login, buf );
        fprintf( rfile, "\nLast Login:                       %s", (st_eq(buf, "-") ? "Unknown (never?)" : buf) );
        if ( ! (st_eq(buf, "-") && lt.last_login_time[0] == 0 && lt.last_login_time[1] == 0 && lt.last_login_time[2] == 0) )
            fprintf( rfile, " @ %02d:%02d:%02d", lt.last_login_time[0], lt.last_login_time[1], lt.last_login_time[2] );

        fprintf( rfile, "\nRestrictions:                     %02Xh", lt.restrictions );

//~~~   GetDiskSpaceUsedByObject()

        fprintf( rfile, "\nDisk Space Utilization:" );
        totalInUse = 0 ;
        LOOP( x, 0, (int) maxVolume, 1 )
        {
            nw_get_obj_restrictions( serverConnID, (BYTE) x, oID, &volRestriction, &volInUse, &isRestr );
            volInUse = volInUse * vi[x].volBlockSize ;
            fprintf( rfile, "\n   %-30s %ld Kbytes", vi[x].volName, volInUse );
            if ( isRestr )
                fprintf( rfile, " (restricted to %ld Kbytes)", volRestriction * (ULONG) vi[x].volBlockSize );
            totalInUse += volInUse ;
        }
        fprintf( rfile, "\n   Total network space in use:    %ld Kbytes", totalInUse );






//      
//      Last Login:                    (Unknown)
//      File Server Console Operator:  No
//      Disk Space In Use:             0 KBytes
//      User ID:                       E5000002
//      

////    {
////    int x ;
////    fprintf( rfile, "\n%34s", "" );
////    for ( x= 0 ; x< 64 ; x++ )
////        fprintf( rfile, "%02x ", propValue[x] );
////    fprintf( rfile, "\n%34s", "" );
////    for ( x=64 ; x< 86 ; x++ )
////        fprintf( rfile, "%02x ", propValue[x] );
////    fprintf( rfile, "~~" );
////    for ( x=86 ; x<128 ; x++ )
////        if ( propValue[x] )
////            fprintf( rfile, " %3d=%02x", propValue[x] );
        fprintf( rfile, "\n" );
////    }
    }
}


void    check_problems( char oName[], WORD oType )
{
    int         ccode ;
    INT32       ymd, date_diff ;
    ULONG       oID ;

    ccode = get_login_control( oName, oType, &oID, propValue, &lt );
    if ( ccode )
    {
        fprintf( sfile, "\n%s *** %04Xh *** ", oName, ccode );
    }
    else
    {
        if ( lt.acct_disabled == 0xFF )
        {
            fprintf( sfile, "\nUser %s has disabled account", oName );
            if ( lt.last_login.d == 0 && lt.last_login.m == 0 && lt.last_login.y == 0 )
                fprintf( sfile, "consider removing" );
        }
        if ( lt.last_login.d == 0 && lt.last_login.m == 0 && lt.last_login.y == 0 )
            fprintf( sfile, "\nUser %s has never logged in", oName );
        else
        {
            ymd = (lt.last_login.y * 10000L) + (lt.last_login.m * 100L) + lt.last_login.d ;
            date_diff = DaysBetween( sys_ymd, ymd );
            if ( date_diff > 30 )
                fprintf( sfile, "\nUser %s has not logged in since %s (%ld days ago)", oName, nw_formatted_bindery_date(&lt.last_login, buf), date_diff );
        }
        if ( lt.pw_expires.y==85 && lt.pw_expires.m==1 && lt.pw_expires.d==1 )
            fprintf( sfile, "\nUser %s has not changed password from given password or password has expired", oName );
        if ( lt.sim_connect.n == 0 )
            fprintf( sfile, "\nUser %s has unlimited simultaneous connections", oName );
        else
        if ( lt.sim_connect.n > 2 )
            fprintf( sfile, "\nUser %s can have %d simultaneous connectionslimit to 1 or 2", oName, lt.sim_connect.n );

        if ( lt.min_pw_length == 0 &&
             lt.days_betw_pw_changes.n == 0 &&
             (lt.pw_expires.y==0 && lt.pw_expires.m==0 && lt.pw_expires.d==0) &&
             (lt.grace_allowed == 0xFF && lt.grace_left == 0xFF) )
            fprintf( sfile, "\nUser %s is not required to have a password", oName );
        else
        {
            if ( lt.days_betw_pw_changes.n > 60 )
                fprintf( sfile, "\nUser %s should change password more frequently than %d days", oName, lt.days_betw_pw_changes.n );
            if ( lt.min_pw_length < 6 )
                if ( lt.min_pw_length == 0 )
                    fprintf( sfile, "\nUser %s has no minimum password length, should be 6 or more", oName, lt.min_pw_length );
                else
                    fprintf( sfile, "\nUser %s has minimum password length of %d characters, should be 6 or more", oName, lt.min_pw_length );
            if ( lt.grace_allowed == 0xFF || lt.grace_left == 0xFF )
                fprintf( sfile, "\nUser %s has unlimited grace logins", oName );
        }
    }
}


void    count_users( void )
{
    if ( write_status )
    {
        sfile = fopen( statusfile, "w" );
        if ( ! sfile )
        {
            write_status = FALSE ;
            pr( 1, 1, "Could not open problem file" );
        }
        else
            fprintf( sfile, "User Audit Status File created on %s %s %s\n", TodaysDay(), TodaysDate(), TodaysTime() );
    }
    if ( research )
    {
        rfile = fopen( researchfile, "w" );
        if ( ! rfile )
        {
            research = FALSE ;
            pr( 1, 1, "Could not open user information file" );
        }
        else
            fprintf( rfile, "User Audit Information File created on %s %s %s\n", TodaysDay(), TodaysDate(), TodaysTime() );
    }

    for ( this_user=0 ; this_user<user_end ; this_user++ )
    {
        total_users += 1 ;
        if ( write_status )
            check_problems( users[this_user], OT_USER );
        if ( research )
            research_login_data( users[this_user], OT_USER );
    }

    if ( research )
        pr( 0, 1, "" );
    fprintf( sfile, "\n\n" );
    fclose( sfile );
    fclose( rfile );
}


void    load_lookuplist( void )
{
    FILE        *flookup ;
    char        *p ;

    StatusMsg( "Loading password check file %s", datafile );
    if ( (flookup = fopen( (const char *) datafile, "r")) == NULL )
        abort_job( 1, exitmsg[1] );
    lookup_end = 0 ;
    while ( (p = fgets( buf1, BUFSIZE, flookup )) )
    {
        StrDetab( (STRING)buf, (STRING)buf1, BUFSIZE );
        st_reduce_spaces( (STRING)buf );
        st_trim( (STRING)buf );
        _strupr( buf );
        if ( lookup_end < MAXLOOK )
        {
            if ( strlen(buf) )
            {
                lookup[lookup_end] = _strdup(buf);
                if ( lookup[lookup_end] == NULL )
                    abort_job( 2, exitmsg[2] );
                lookup_end += 1 ;
            }
        }
        else
        {
            pr( 0, 1, "Too many entries in the lookup list: %s", buf );
        }
    }
    fclose( flookup );
    pr( 0, 1, "%d Passwords to check", lookup_end );
}


void    load_userlist( void )
{
    int         ret ;
    char        objectName[B_USERNAME_SIZE], objectNameToSearch[B_USERNAME_SIZE] ;
    long        objectID ;
    WORD        objectType, objectTypeToSearch ;
    BYTE        objectHasProperties, objectFlag, objectSecurity ;

    StatusMsg( "Loading user list" );
    user_end = 0 ;

    for ( objectID = -1, ret = 0 ; (! ret) ; )  // check all users
    {
        strcpy( objectNameToSearch, "*" );
        objectTypeToSearch = OT_USER ;
        ret = ScanBinderyObject( objectNameToSearch, objectTypeToSearch,
                                 &objectID, objectName, &objectType,
                                 &objectHasProperties, &objectFlag,
                                 &objectSecurity );
        if ( ! ret )
        {
            strcpy( buf, objectName );
            st_reduce_spaces( (STRING)buf );
            st_trim( (STRING)buf );
            _strupr( buf );
            if ( user_end < MAXUSER )
            {
                if ( strlen(buf) )
                {
                    users[user_end] = _strdup(buf);
                    if ( users[user_end] == NULL )
                        abort_job( 5, exitmsg[5] );
                    user_end += 1 ;
                }
            }
            else
            {
                pr( 0, 1, "Too many entries in the user list: %s", buf );
            }
        }
    }
    qsort( (char *)users, (size_t)user_end, (size_t)sizeof(char *), qs_compare );
}


void    intruder_lockout_off( void )
{
}


void    check_results( int vcode, STRING user )
{
    if ( vcode == SERVER_OUT_OF_MEMORY )        // 0x96
    {
        upd_statusfile( "Server out of memory.  Aborted at user %s.", user );
        abort_job( 3, exitmsg[3] );
    }
    if ( vcode == SERVER_BINDERY_LOCKED )       // 0xFE
    {
        upd_statusfile( "Bindery locked.  User %s not (completely) checked!", user );
        tm_sleep_secs( 600 );                   // sleep 10 minutes
    }
    if ( vcode == INTRUDER_DETECTION_LOCK )     // 0xC5
    {
        sprintf( buf, "From %s: Intruder detection caused by Password Audit", userName );
        BroadcastToConsole( buf );
        abort_job( 4, exitmsg[4] );
    }
}


void    do_parm( STRING par )
{
    UCHAR       p ;

    _strupr( par );
    p = *par++ ;
    switch ( p )
    {
        case 'L' :                          // set users with offending passwords to "locked out" status (disable account)
            disable_account = TRUE ;
            break ;
        case 'N' :                          // check non-users (! OT_USER)
            break ;
        case 'B' :                          // begin checking at name
            if ( *par == '=' )
            {
                par++ ;
                if ( strlen(par) )
                {
                    ST_CPY ( startName, par );
                    st_trim( startName );
                    restart = TRUE ;
                    pr( 0, 1, "Starting to check at %s", startName );
                }
            }
            break ;
        case 'S' :                          // show offending passwords
            show_pw = TRUE ;
            break ;
        case 'D' :                          // check only passwords changed since date
            break ;
        case 'U' :                          // unsorted
            fl_unsorted = TRUE ;
            break ;
        case '?' :                          // help
        case 'H' :
            do_help();
        default  :
            fprintf( stderr, "\n\n\nSwitch \"%c\" is not legal.\a", p );
            do_help();
    }
}


void    status( STRING nm, STRING pw, int pwx )
{
/*
User supervisor nnn of nnn: nnn%   xxxxxxxxxxxxxxxxxxxxxxxxxx nnnn of nnnn: nnn%
*/
    static UCHAR l_nm[B_USERNAME_SIZE]="", mbuf[51] ;
    static ULONG lPct ;
    static int   pct ;

    DspSysTime();
    if ( ! st_eq(nm,l_nm) )
    {
        lPct = (this_user+1) * 100L ;
        lPct /= total_users ;
        pct = (int) lPct ;
        ST_CPY( l_nm, nm );
        StatusMsg( "" );
        sprintf( ErrMsg_buf_, "%-.17s  %d of %d: %d%%", l_nm, this_user+1, total_users, pct );
        DspMsg( vid.rows, 1, vid.sts_fg, vid.sts_bg, ErrMsg_buf_, NO );
    }
    lPct = (pwx+1) * 100L ;
    lPct /= lookup_end ;
    pct = (int) lPct ;
#if IS_USED_INTERNAL_ONLY
    sprintf( ErrMsg_buf_, "%26.26s  %d of %d: %d%%", pw, (pwx+1), lookup_end, pct );
#else
    sprintf( ErrMsg_buf_, "%32s: %d%%", "Words processed", pct );
#endif
    sprintf( mbuf, "%45.45s", ErrMsg_buf_ );
    DspMsg( vid.rows, 36, vid.sts_fg, vid.sts_bg, mbuf, NO );
}


void    process_keystroke( STRING obj, int index )
{
    int         ret, resp=1 ;
    STRING      resp_str ;
    UCHAR       *save=NULPTR ;

    FlushKeyboard();                            // prevent type-ahead
    save = SaveScreen( save );                  // save the screen
    NewScreen();
    ret = sc_menu_init( UNDEF, UNDEF, POP_MENU, "Choose Action", 2, NO, YES );
    if ( ret != OK )
        abort_job( 9, "menu init error" );

    if ( keyboard_locked )
    {
        ret = sc_menu_item( 0, 0, "Unlock", "Unlock the keyboard" );
        if ( ret != OK )
            abort_job( 9, "menu choice error" );
    }
    else
    {
        ret = sc_menu_item( 0, 0, "Lock", "Lock the keyboard" );
        if ( ret != OK )
            abort_job( 9, "menu choice error" );
        ret = sc_menu_item( 0, 0, "Abort", "Stop processing" );
        if ( ret != OK )
            abort_job( 9, "menu choice error" );
    }
    resp = sc_menu_choice( resp, &resp_str );
    save = RestoreScreen( save );               // put the screen back first
    if ( resp == ERROR )
        abort_job( 9, "menu error" );
    if ( resp == 0 )
        return ;
    if ( st_eq(resp_str,"abort") )
    {
        upd_statusfile( "Aborted at %s[%s]", obj, lookup[index] );
        topbox( "Process aborted by user" );
        MoveCursor( vid.rows-1, 1 );
        abort_job( GD_ERROR_NONE, "" );
    }
    else
        enter_kb_pw();                          // password stuff
}


void    found_problem( int which, STRING Uname, STRING Upw )
{
    risks += 1 ;
    sprintf( buf, "From %s: Password Audit on %s %s %s", userName, TodaysDay(), TodaysDate(), TodaysTime() );
    BroadcastToConsole( buf );
    switch ( which )
    {
        case 1 :
            sprintf( buf1, "Password of user %s is loginID", Uname );
            sprintf( buf , "! %s password of user %s is same as login id", get_global_program_name(), Uname );
            break ;
        case 2 :
            sprintf( buf1, "User %s has no password", Uname );
            sprintf( buf , "! %s user %s has no password", get_global_program_name(), Uname );
            break ;
        case 3 :
            sprintf( buf1, "Guessed password of user %s", Uname );
            sprintf( buf , "! %s guessed password of user %s", get_global_program_name(), Uname );
            break ;
    }
    nw_logmsg( buf );
    pr( 0, 1, "%s", buf1 );
    upd_statusfile( "-> %s <-", buf1 );
    sprintf( buf, "From %s: %s", userName, buf1 );
    BroadcastToConsole( buf );
    if ( show_pw )
        append_to_secret_file( Uname, Upw );
}


void    process_user( STRING uN, WORD objectType )
{
    static char uName[B_USERNAME_SIZE] ;
    int         x, verify, ccode ;
    ULONG       oID ;
    INT32       exp_ymdn ;

    DspSysDate();
    strcpy( uName, uN );
    if ( restart )
    {
        if ( st_eq(uName, startName) )
            restart = FALSE ;
        else
            pr( 0, 1, "Bypassing %s", uName );
    }
    if ( ! restart )
    {
        if ( st_eq(uName, userName) )  // should know my own password...
        {
            pr( 0, 1, "Skipping %s", uName );
        }
        else
        {
            ccode = get_login_control( uName, objectType, &oID, propValue, &lt );
            if ( ! ccode )
            {
                if ( lt.acct_disabled == 0xFF )
                {
                    sprintf( buf, "Bypassing disabled account of %s", uName );
                    upd_statusfile( "%s", buf );
                    pr( 0, 1, "%s", buf );
                    return ;
                }
                exp_ymdn = Date2Num( nw_bindery_date_to_ymd(&lt.pw_expires) );
                nw_formatted_bindery_date(&lt.pw_expires,buf) ;
                if ( (exp_ymdn <= sys_ymdn) && (! st_eq(buf,"-")) )
                {
                    upd_statusfile( "Password of %s expired on %s (%ld days ago)--not checking", uName, buf, (sys_ymdn - exp_ymdn) );
                    pr( 0, 1, "Bypassing account of %s--password expired on %s", uName, buf );
                    return ;
                }
            }

            upd_statusfile( "Checking account of %s", uName );
            status( uName, uName, 0 );
            verify = VerifyBinderyObjectPassword( uName, objectType, uName );
            if ( ! verify )
            {
                found_problem( 1, uName, uName );
            }
            else
            {
                status( uName, "<none>", 0 );
                verify = VerifyBinderyObjectPassword( uName, objectType, "" );
                if ( ! verify )
                    found_problem( 2, uName, "[no_password]" );
            }
            if ( verify )
            {
                check_results( verify, uName );
                progress = (ULONG) lookup_end * ((ULONG) this_user) ;
                for ( x = 0 ; x < lookup_end ; x++ )
                {
                    status( uName, lookup[x], x );
                    progress += 1 ;
                    verify = VerifyBinderyObjectPassword( uName, objectType, lookup[x] );
                    if ( verify )
                    {
                        check_results( verify, uName );
                    }
                    else
                    {
                        found_problem( 3, uName, lookup[x] );
                        x = lookup_end ;
                    }
//////////////////////////////~~~~~~~~~~~~~~~~
                    if ( kb_Hit() )
                        process_keystroke( uName, x );
                }
            }
        }
    }
}


void    process_sorted( void )
{
    for ( this_user=0 ; this_user<user_end ; this_user++ )
    {
        process_user( users[this_user], OT_USER );
    }
}


void    process_unsorted( void )
{
    int         ret ;
    char        objectName[B_USERNAME_SIZE], objectNameToSearch[B_USERNAME_SIZE] ;
    long        objectID ;
    WORD        objectType, objectTypeToSearch ;
    BYTE        objectHasProperties, objectFlag, objectSecurity ;

    for ( objectID=-1, ret=0, this_user=0 ; (! ret) ; ) // check all users
    {
        strcpy( objectNameToSearch, "*" );
        objectTypeToSearch = OT_USER ;
        ret = ScanBinderyObject( objectNameToSearch, objectTypeToSearch,
                                 &objectID, objectName, &objectType,
                                 &objectHasProperties, &objectFlag,
                                 &objectSecurity );
        if ( ! ret )
        {
            process_user( objectName, objectType );
            this_user += 1 ;
        }
    }
}


void    main_exit( void )
{
    if ( finish_off_code == BAD_EXIT )
        pr( 1, 1, "Aborting program %s", get_global_program_name() );
    kb_idle_remove();
    fcloseall();
    CursOn();
}


void    main_init( int argc, char *argv[] )
{
    UCHAR       *ptr, parm[256] ;
    int         i ;
    INT16       sys_y, sys_m, sys_d ;

    fs_init( FALSE, SHAREWARE_VERSION, 0, pgVERS_, pgTITL_, pgCOPR_ );
    Allow_No_Break();                           // allow no ^Break to abort
    atexit( main_exit );
    NewScreen();

    if ( ! nw_installed_NetWare() )
        abort_job( 9, "Not logged in to a NetWare file server!" );

    GetFileServerName( (WORD) 0, serverName );
    GetConnectionID( serverName, &serverConnID );
    nw_get_username( userName );

    if ( ! nw_is_supervisor(userName) )
        abort_job( 9, super_rights );

    sprintf( buf, "%s [%s:%s]", pgm_title, serverName, userName );
    SetGlobalString( PGM_TITLE, buf );
    NewScreen();
    ClearRect( 3,1, 4,vid.cols, vid.err_fg, vid.err_bg );
    pr_centered( 3, pgm_copyrite );
//  pr_centered( 4, pgm_rights   );
    topbox( "Press a key for menu" );

    ST_CPY( datafile, argv[0] );
    ptr = strrchr(datafile, '.');
    strcpy( ptr, ".DAT" );

    dt_getsystemdate( &sys_ymd, &sys_y, &sys_m, &sys_d );
    sys_ymdn = Date2Num( sys_ymd );

    argc -= 1 ;
    LOOP( i, 1, argc, 1 )                       // check parameters
    {
        ST_CPY( parm, argv[i] );
        if ( parm[0] == '/' || parm[0] == '-' )
            do_parm( &parm[1] );
    }

    if ( sizeof(lt) != 128 )
    {
        pr( 1, 1, "Actual size of LT: %d", sizeof(lt) );
        abort_job( 9, "LT structure size error" );
    }
}


void    init_volume_info( void )
{
    int         x ;
    WORD        ccode ;

    LOOP( x, 0, 31, 1 )
    {
        vi[x].volName = NULPTR ;
        vi[x].volNumber = 0 ;
        vi[x].volBlockSize = 0 ;
        ccode = GetVolUsage( serverConnID, (WORD) x, &vu );
        if ( ! ccode )
        {
            if ( ! st_eq(vu.volName," ") )
            {
                vi[x].volName = strdup( vu.volName );
                vi[x].volNumber = x ;
                vi[x].volBlockSize = (int) (vu.blockSize / SZ_kBYTE);
                maxVolume = x ;
            }
        }
    }
}


void    main( int argc, char *argv[] )
{
    main_init( argc, argv );

    if ( ! YesNo( "Is INTRUDER LOCKOUT TURNED OFF", NO, TRUE ) )
        happy_ending();

    intruder_lockout_off();
    init_volume_info();
    load_lookuplist();
    load_userlist();
    count_users();
    checks = ((ULONG) lookup_end) * total_users ;
    if ( checks > 200 )
        pr( 0, 1, "This may take a while.  We need to check %ld combinations!", checks );

    kb_idle_reset( 60 );                        // blank after 1 minute
    kb_idle_install( kb_idle_time_n_blank );
    blank_message = "Screen Saver" ;

    if ( fl_unsorted )
        process_unsorted();
    else
        process_sorted();

    while ( keyboard_locked )
        enter_kb_pw();

    if ( risks )
        pr( 1, 1, "%d user%s with a problem password", risks, risks==1?"":"s" );

    pr( 1, 1, "Make sure to turn intruder lockout back on!" );
    sn_beep();

    happy_ending();
}
